# coding=utf-8
from __future__ import absolute_import, print_function

import itertools

from suanpan.arguments import Int, String
from suanpan.proxy import Proxy
from suanpan import error

# try:
#     from suanpan.dw import hive, mysql, postgres  # odps
# except ImportError:
#     from suanpan.dw import mock
#     hive = mysql = postgres = mock  # odps


class DataWarehouseProxy(Proxy):
    # MAPPING = {
    #     "hive": hive.DataWarehouse,
    #     # "odps": odps.DataWarehouse,
    #     "postgres": postgres.DataWarehouse,
    #     "mysql": mysql.DataWarehouse,
    # }
    DEFAULT_ARGUMENTS = [String("dw-type", default="odps")]
    HIVE_ARGUMENTS = [
        String("dw-hive-host", default="localhost"),
        Int("dw-hive-port", default=10000),
        String("dw-hive-database", default="default"),
        String("dw-hive-username"),
        String("dw-hive-password"),
        String("dw-hive-auth"),
    ]

    ODPS_ARGUMENTS = [
        String("dw-odps-access-id"),
        String("dw-odps-access-key"),
        String(
            "dw-odps-endpoint", default="http://service.cn.maxcompute.aliyun.com/api"
        ),
        String("dw-odps-project"),
    ]

    POSTGRES_ARGUMENTS = [
        String("dw-postgres-host", default="localhost"),
        Int("dw-postgres-port", default=5432),
        String("dw-postgres-database"),
        String("dw-postgres-username"),
        String("dw-postgres-password"),
    ]

    MYSQL_ARGUMENTS = [
        String("dw-mysql-host", default="localhost"),
        Int("dw-mysql-port", default=3306),
        String("dw-mysql-database"),
        String("dw-mysql-username"),
        String("dw-mysql-password"),
    ]

    ARGUMENTS = list(
        itertools.chain(
            DEFAULT_ARGUMENTS,
            HIVE_ARGUMENTS,
            ODPS_ARGUMENTS,
            POSTGRES_ARGUMENTS,
            MYSQL_ARGUMENTS,
        )
    )

    def importBackend(self, backendType):
        try:
            if backendType == "hive":
                from suanpan.dw import hive
                return hive.DataWarehouse
            elif backendType == "postgres":
                from suanpan.dw import postgres
                return postgres.DataWarehouse
            elif backendType == "mysql":
                from suanpan.dw import mysql
                return mysql.DataWarehouse
            else:
                raise error.ProxyError(
                    f"{self.name} don't supported backend type {backendType}"
                )
        except ImportError:
            from suanpan.dw import mock
            return mock.DataWarehouse


dw = DataWarehouseProxy()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import g
from suanpan.proxy import Proxy

TYPE = g.screenshotsType
PATTERN = g.screenshotsPattern
STORAGE_KEY = g.screenshotsStorageKey
THUMBNAIL_STORAGE_KEY = g.screenshotsThumbnailStorageKey


class Screenshots(Proxy):
    MAPPING = {
        "index": "suanpan.screenshots.base.ScreenshotsIndexSaver",
        "time": "suanpan.screenshots.base.ScreenshotsTimeSaver",
    }
    DEFAULT_PATTERN_MAPPING = {"index": "{index}.png", "time": "{time}.png"}

    def __init__(self, *args, **kwargs):
        kwargs.setdefault(self.TYPE_KEY, TYPE)
        kwargs.setdefault("name", STORAGE_KEY)
        kwargs.setdefault("thumbnail", THUMBNAIL_STORAGE_KEY)
        kwargs.setdefault(
            "pattern", PATTERN or self.DEFAULT_PATTERN_MAPPING.get(kwargs["type"])
        )
        super().__init__(*args, **kwargs)


screenshots = Screenshots()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import objects
from suanpan.mstorage import mstorage
from suanpan.stream import vars


class Context(objects.Context):
    @property
    def vars(self):
        return vars.Variables(self, mstorage)



from suanpan.stypes.app import AppT
from suanpan.stypes.trigger import TriggerT
from suanpan.stypes.node import NodeT
from suanpan.stypes.storage import StorageT
from suanpan.stypes.argument import ArgumentT



import sys
import inspect
from suanpan.app import app, MessageHandler
from suanpan.node import node


def simple_run():
    run_func = None
    initialize_func = None

    main = sys.modules['__main__']
    for name, func in inspect.getmembers(main):
        if name == 'run':
            run_func = func
        elif name == 'initialize':
            initialize_func = func

    if run_func:
        app.use(MethodHandler(run_func, initialize_func))
        app.start()


class MethodHandler(MessageHandler):
    def __init__(self, run_f, initialize_f=None):
        super().__init__()
        self._run_f = run_f
        self._initialize_f = initialize_f

    @property
    def inputs(self):
        return node.inargs

    @property
    def outputs(self):
        return node.outargs

    @property
    def params(self):
        return node.paramargs

    def initialize(self):
        if self._initialize_f:
            self._initialize_f(self.context)

    def run(self, context):
        # inputs = app.load(node.inargs)
        return self._run_f(context)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.api import app
from suanpan.utils import json


class Interface(object):
    def __init__(self, request):
        super().__init__()
        self.request = request

    def create(self, appId):
        return self.request.post(
            "/api/app", json={"id": appId, "graph": json.dumps(app.getAppGraph(appId))}
        )

    def delete(self, appId):
        return self.request.delete(f"/api/app/{appId}")



import sys
import types
import importlib.util
import importlib.abc


class _LazyModule(types.ModuleType):
    """A subclass of the module type which triggers loading upon attribute access."""

    def __getattribute__(self, attr):
        if attr in ('__name__',):
            return super(_LazyModule, self).__getattribute__(attr)

        print("reload lazy_imported '{}' with attr '{}'".format(self.__name__, attr))
        """Trigger the load of the module and return the attribute."""
        # All module metadata must be garnered from __spec__ in order to avoid
        # using mutated values.
        # Stop triggering this method.
        self.__class__ = types.ModuleType
        # Get the original name to make sure no object substitution occurred
        # in sys.modules.
        original_name = self.__spec__.name
        # Figure out exactly what attributes were mutated between the creation
        # of the module and now.
        attrs_then = self.__spec__.loader_state['__dict__']
        original_type = self.__spec__.loader_state['__class__']
        attrs_now = self.__dict__
        attrs_updated = {}
        for key, value in attrs_now.items():
            # Code that set the attribute may have kept a reference to the
            # assigned object, making identity more important than equality.
            if key not in attrs_then:
                attrs_updated[key] = value
            elif id(attrs_now[key]) != id(attrs_then[key]):
                attrs_updated[key] = value
        self.__spec__.loader.exec_module(self)
        # If exec_module() was used directly there is no guarantee the module
        # object was put into sys.modules.
        if original_name in sys.modules:
            if id(self) != id(sys.modules[original_name]):
                raise ValueError(f"module object for {original_name!r} "
                                 "substituted in sys.modules during a lazy "
                                 "load")
        # Update after loading since that's what would happen in an eager
        # loading situation.
        self.__dict__.update(attrs_updated)
        return getattr(self, attr)

    def __delattr__(self, attr):
        """Trigger the load and then perform the deletion."""
        # To trigger the load and raise an exception if the attribute
        # doesn't exist.
        self.__getattribute__(attr)
        delattr(self, attr)

    def __repr__(self):
        return "<module '{}' from 'lazy_import'>".format(self.__name__)


class LazyLoader(importlib.abc.Loader):
    """A loader that creates a module which defers loading until attribute access."""

    @staticmethod
    def __check_eager_loader(loader):
        if not hasattr(loader, 'exec_module'):
            raise TypeError('loader must define exec_module()')

    @classmethod
    def factory(cls, loader):
        """Construct a callable which returns the eager loader made lazy."""
        cls.__check_eager_loader(loader)
        return lambda *args, **kwargs: cls(loader(*args, **kwargs))

    def __init__(self, loader):
        self.__check_eager_loader(loader)
        self.loader = loader

    def create_module(self, spec):
        return self.loader.create_module(spec)

    def exec_module(self, module):
        """Make the module load lazily."""
        module.__spec__.loader = self.loader
        module.__loader__ = self.loader
        # Don't need to worry about deep-copying as trying to set an attribute
        # on an object would have triggered the load,
        # e.g. ``module.__spec__.loader = None`` would trigger a load from
        # trying to access module.__spec__.
        loader_state = {'__dict__': module.__dict__.copy(), '__class__': module.__class__}
        module.__spec__.loader_state = loader_state
        module.__class__ = _LazyModule


def lazy_import(name):
    try:
        return sys.modules[name]
    except KeyError:
        spec = importlib.util.find_spec(name)
        loader = LazyLoader(spec.loader)
        spec.loader = loader
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module
        loader.exec_module(module)
        return module



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import g
from suanpan.api.requests import affinity


def getAppGraph(appId):
    return affinity.get(f"/app/graph/{appId}")["data"]


def updateAppGraph(appId, graph):
    return affinity.post(f"/app/graph/{appId}", json={"graph": graph})


def revertAppGraph(appId):
    return affinity.post(f"/app/graph/{appId}/revert")


def checkAppStatus(appId=None):
    return affinity.post("/app/status", json={"id": appId or g.appId})


# mode: runStart, runStop, runStartStop
def runApp(nodeId, appId=None, mode="runStart"):
    return affinity.post(
        "/app/run", json={"id": appId or g.appId, "nodeId": nodeId, "type": mode}
    )


def startAppCron(
    period, mode, startEffectTime, endEffectTime, nodeId, config, appId=None
):
    return affinity.post(
        "/app/cron/start",
        json={
            "id": appId or g.appId,
            "cronConfig": {
                "period": period,
                "runType": mode,
                "startEffectTime": startEffectTime,
                "endEffectTime": endEffectTime,
                "nodeId": nodeId,
                "config": config,
            },
        },
    )


def stopAppCron(appId=None):
    return affinity.post("/app/cron/stop", json={"id": appId or g.appId})


def createApp(appName, appType, dirId=None):
    return affinity.post(
        "/app/create", json={"name": appName, "dir": dirId, "type": appType}
    )


def deleteApp(appId=None):
    return affinity.post("/app/del", json={"id": appId or g.appId})


def copyApp(name, appType, appId=None, withData=False, description=None, dirId=None):
    return affinity.post(
        "/app/duplicate",
        json={
            "id": appId or g.appId,
            "name": name,
            "dir": dirId,
            "type": appType,
            "withData": withData,
            "description": description,
        },
    )


def listApp(appType, withStatus=False, limit=9999):
    return affinity.post(
        "/app/list", json={"limit": limit, "type": appType, "withStatus": withStatus}
    )



import abc


class NodeT(object):
    @property
    @abc.abstractmethod
    def info(self):
        """the `info` of the node"""
        ...

    @property
    @abc.abstractmethod
    def inputs(self):
        """the `inputs` of the node"""
        ...

    @property
    @abc.abstractmethod
    def ins(self):
        """the `inputs` of the node, alias of `inputs`"""
        ...

    @property
    @abc.abstractmethod
    def outputs(self):
        """the `outputs` of the node"""
        ...

    @property
    @abc.abstractmethod
    def outs(self):
        """the `outputs` of the node, alias of `outputs`"""
        ...

    @property
    @abc.abstractmethod
    def params(self):
        """the `params` of the node"""
        ...

    @property
    @abc.abstractmethod
    def inargs(self):
        """a list of inputs"""
        ...

    @property
    @abc.abstractmethod
    def outargs(self):
        """a list of outputs"""
        ...

    @abc.abstractmethod
    def json(self):
        """json object of node"""
        ...



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import g


class Interface(object):
    def __init__(self, request):
        super().__init__()
        self.request = request

    def init(self, suanpanUrl, userId=None, accessKey=None, accessSecret=None):
        return self.request.post(
            "/api/suanpan/init",
            json={
                "suanpanUrl": suanpanUrl,
                "userId": userId or g.userId,
                "accessKey": accessKey or g.accessKey,
                "accessSecret": accessSecret or g.accessSecret,
            },
        )



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.app import app
from suanpan.app.modules.base import Module
from suanpan.utils import image

module = Module()


@module.fn
def saveDebugImageAndNotify(img, itype):
    saveDebugImage(img, itype)
    notifyImageMeta({"type": itype})


@module.fn
def saveDebugImage(img, itype):
    image.save(app.sio.static("debug", f"{itype}.jpg"), img)


@module.fn
def notifyImageMeta(meta):
    app.sio.emit("images.meta.notify", meta)


@module.on("images.meta.get")
def getImageMeta(context):
    args = context.args
    return {
        "type": args.type,
        "src": f"debug/{args.type}.jpg",
    }



import abc
import typing as t

if t.TYPE_CHECKING:
    from suanpan.stypes.app import AppT


class IntervalT(object):
    @abc.abstractmethod
    def __init__(self, seconds, pre=False):
        """init interval instance

        Arguments:
            seconds: the seconds this interval wait
            pre: if True, wait before doing the trigger function
                else wait after doing the trigger function

        Example:
            interval = Interval(3600)
        """
        ...


class TriggerT(AppT):
    @abc.abstractmethod
    def loop(self, interval: IntervalT):
        """run the trigger every intervals"""
        ...



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import app
from suanpan import arguments as base  # pylint: disable=unused-import
from suanpan.arguments import auto  # pylint: disable=unused-import
from suanpan.dw import arguments as dw
from suanpan.imports import imports
from suanpan.storage import arguments as storage

Int = imports(f"suanpan.{app.TYPE}.arguments.Int")
Float = imports(f"suanpan.{app.TYPE}.arguments.Float")
Bool = imports(f"suanpan.{app.TYPE}.arguments.Bool")
String = imports(f"suanpan.{app.TYPE}.arguments.String")
List = imports(f"suanpan.{app.TYPE}.arguments.List")
ListOfString = imports(f"suanpan.{app.TYPE}.arguments.ListOfString")
ListOfInt = imports(f"suanpan.{app.TYPE}.arguments.ListOfInt")
ListOfFloat = imports(f"suanpan.{app.TYPE}.arguments.ListOfFloat")
ListOfBool = imports(f"suanpan.{app.TYPE}.arguments.ListOfBool")
IntOrFloat = imports(f"suanpan.{app.TYPE}.arguments.IntOrFloat")
IntFloatOrString = imports(f"suanpan.{app.TYPE}.arguments.IntFloatOrString")
BoolOrString = imports(f"suanpan.{app.TYPE}.arguments.BoolOrString")
StringOrListOfFloat = imports(f"suanpan.{app.TYPE}.arguments.StringOrListOfFloat")
Json = imports(f"suanpan.{app.TYPE}.arguments.Json")

NameFile = imports(f"suanpan.{app.TYPE}.arguments.NameFile")
File = imports(f"suanpan.{app.TYPE}.arguments.File")
Folder = imports(f"suanpan.{app.TYPE}.arguments.Folder")
Data = imports(f"suanpan.{app.TYPE}.arguments.Data")
Csv = imports(f"suanpan.{app.TYPE}.arguments.Csv")
Excel = imports(f"suanpan.{app.TYPE}.arguments.Excel")
Npy = imports(f"suanpan.{app.TYPE}.arguments.Npy")
Visual = imports(f"suanpan.{app.TYPE}.arguments.Visual")
Model = imports(f"suanpan.{app.TYPE}.arguments.Model")
H5Model = imports(f"suanpan.{app.TYPE}.arguments.H5Model")
Checkpoint = imports(f"suanpan.{app.TYPE}.arguments.Checkpoint")
JsonModel = imports(f"suanpan.{app.TYPE}.arguments.JsonModel")
Image = imports(f"suanpan.{app.TYPE}.arguments.Image")

Table = imports(f"suanpan.{app.TYPE}.arguments.Table")
DataFrame = imports(f"suanpan.{app.TYPE}.arguments.DataFrame")



# coding=utf-8
from __future__ import absolute_import, print_function
import sys
import datetime
import functools
import json
from suanpan import path
from suanpan.lazy_import import lazy_import

if sys.platform == "win32":
    import numpy as np
    import pandas as pd
else:
    np = lazy_import('numpy')
    pd = lazy_import('pandas')


class SPEncoder(json.JSONEncoder):
    INNER_DATETIME = (datetime.datetime,)

    def default(self, obj):  # pylint: disable=method-hidden
        if isinstance(obj, self.INNER_DATETIME):
            return obj.replace(tzinfo=datetime.timezone.utc).isoformat()
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.generic):
            return obj.item()
        if isinstance(obj, pd.DataFrame):
            return obj.to_dict()
        return json.JSONEncoder.default(self, obj)


if sys.version_info.minor < 9:
    _load = functools.partial(json.load, encoding="utf-8")
    _loads = functools.partial(json.loads, encoding="utf-8")
else:
    _load = functools.partial(json.load)
    _loads = functools.partial(json.loads)
_dump = functools.partial(json.dump, ensure_ascii=False, cls=SPEncoder)
_dumps = functools.partial(json.dumps, ensure_ascii=False, cls=SPEncoder)


def _loadf(file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    with open(file, "r", encoding=encoding) as _file:
        return _load(_file, *args, **kwargs)


def _dumpf(s, file, *args, **kwargs):
    encoding = kwargs.pop("encoding", "utf-8")
    path.safeMkdirsForFile(file)
    with open(file, "w", encoding=encoding) as _file:
        return _dump(s, _file, *args, **kwargs)


def load(file, *args, **kwargs):
    _l = _loadf if isinstance(file, str) else _load
    return _l(file, *args, **kwargs)


def dump(s, file, *args, **kwargs):
    _d = _dumpf if isinstance(file, str) else _dump
    return _d(s, file, *args, **kwargs)


loads = _loads
dumps = _dumps



# coding=utf-8
from __future__ import absolute_import, print_function

import zmq

from suanpan.device.base import BaseDevice


class ZMQDevice(BaseDevice):
    def __init__(self, *args, **kwargs):
        self._context = None
        self._socket = None
        super().__init__(*args, **kwargs)

    def connect(self):
        self._context = zmq.Context.instance()
        self._socket = self._context.socket(zmq.DEALER)  # pylint: disable=no-member
        self._socket.setsockopt(zmq.IDENTITY, self.name.encode())  # pylint: disable=no-member
        self._socket.connect(self.url)
        return self._socket

    def close(self):
        if self._context:
            self._context.term()
        if self._socket:
            self._socket.close()

    def _send(self, data, target):
        self._socket.send_multipart([data, target])

    def _recv(self):
        return self._socket.recv_multipart()
